import ij.*;
import ij.plugin.*;
import ij.process.*;
import ij.gui.*;
import ij.util.Tools;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import ij.measure.*;
import java.awt.Rectangle;

/**
   This plugin continuously generates Frame-Dimension profile plots as a selection 
   is moved or resized through the XY Coordinate Plane.

   @author Fred Damen <fred@damen.org>

   Version History:
   2020-04-01: Released
 */

public class F_Profiler implements PlugIn, 
                                    MouseListener, 
                                    MouseMotionListener, 
                                    MouseWheelListener, 
                                    Measurements, 
                                    KeyListener, 
                                    WindowListener {
   ImagePlus img;
   PlotWindow pwin;
   Plot prevPlot = null;
   public double[] x;
   public double[] y;
   public double[] ye;
   String xLabel;
   String yLabel;

   public void run(String arg) {
      img = IJ.getImage();
      int nf = img.getNFrames();
      if (nf<2) {
         IJ.showMessage("Dynamic F-Dimension Profiler", "This command requires a HyperStack.");
         return;
         }

      on();
      img.getWindow().addWindowListener(this);
      engaged = true;
      IJ.showStatus("F-Dimension Profile Engaged: "+img.getTitle());

      x  = new double[nf];
      y  = new double[nf];
      ye = new double[nf];
      String fu = img.getCalibration().getTimeUnit(); 
      try {
         ImageStack is = img.getStack();
         for(int f=0; f<nf; f++) {
            String[] strarr = is.getSliceLabel(img.getStackIndex(1,1,f+1)).split("\n",2)[0].split(",|;",2)[0].split(" = |=| ",2);
            xLabel = strarr[0]+(fu!="" ? " ("+fu+")" : "");
            x[f] = Float.valueOf(strarr[1]).floatValue();
            }
         }
      catch(Throwable e) {
         xLabel = "frame"+(fu!="" ? " ("+fu+")" : "");
         for(int f=0; f<nf; f++)
            x[f] = f;
         }

      yLabel = img.getTitle()+" ("+img.getCalibration().getValueUnit()+")";
      if (!updateProfile())
         return;
      positionPlotWindow();
      }

   private void on() {
      img.getCanvas().addMouseListener(this);
      img.getCanvas().addMouseMotionListener(this);
      img.getCanvas().addKeyListener(this);
      img.getWindow().addMouseWheelListener(this);
      }
   private void off() {
      img.getCanvas().removeMouseListener(this);
      img.getCanvas().removeMouseMotionListener(this);
      img.getCanvas().removeKeyListener(this);
      img.getWindow().removeMouseWheelListener(this);
      }
    boolean engaged = false;
    void disengage() {
       if (!engaged) return;
       if (img.getWindow() != null) {
          off();
          img.getWindow().removeWindowListener(this);
          }
       pwin = null;
       engaged = false;
       IJ.showStatus("F-Dimension Profile Disengaged");
       }

   private boolean updateProfile() {
     Roi roi = img.getRoi();
     if ((img == null) || !img.isVisible() || (pwin != null) && (!pwin.isVisible())) {
         IJ.log("F_Profiler: should not have reached here '"+pwin+"'");
         engaged = true;
         disengage();
         return false;
         }   
      else if (roi == null) {
         Point here = img.getCanvas().getCursorLoc();
         return updateProfile(here.x, here.y);
         }
      else if ((roi.getType() == Roi.POINT) || (roi.getBounds().width==0 && roi.getBounds().height ==0)) {
         Rectangle bounds = roi.getBounds();
         return updateProfile(bounds.x, bounds.y);
         }
      
      else return updateProfile(roi);
      }

   private boolean updateProfile(Roi roi) {
      int nf = img.getNFrames();
      int cs = img.getZ();
      double[] yM = new double[nf];
      double[] ym = new double[nf];
      double[] ys = new double[nf];
      ImageStack is = img.getStack();
      Calibration cal = img.getCalibration();
      for(int f=0; f<nf; f++) {
         ImageProcessor ip = is.getProcessor(img.getStackIndex(1,cs,f+1));
         ip.setRoi(roi);
         ImageStatistics stats = ImageStatistics.getStatistics(ip, MEAN+STD_DEV+MIN_MAX+MEDIAN, cal);
         y[f]  = stats.mean;
         ye[f] = stats.stdDev;
         yM[f] = stats.max;
         ym[f] = stats.min;
         ys[f] = stats.median;
         }
 
      Plot plot = new Plot("Frame Profile ("+img.getTitle()+")", xLabel, yLabel);
      plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
      plot.setXLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
      plot.setYLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));

      if (prevPlot != null) plot.useTemplate(prevPlot, Plot.COPY_SIZE);

      plot.setColor(Color.blue);
      plot.setLineWidth(1);
      plot.addPoints(x,y,ye,Plot.X);
      plot.setColor(Color.red);
      plot.setLineWidth(4);
      plot.addPoints(x,y,Plot.X);

      plot.setColor(Color.green);
      plot.setLineWidth(2);
      plot.addPoints(x,ym,Plot.BOX);
      plot.addPoints(x,yM,Plot.BOX);
      plot.setColor(Color.black);
      plot.addPoints(x,ys,Plot.CIRCLE);
      plot.setLegend("stddev\nmean\nmax\nmin\nmedian",Plot.AUTO_POSITION);
      if (pwin==null) {
         pwin = plot.show();
         pwin.addWindowListener(this);
         }
      else pwin.drawPlot(plot);
      plot.setLimitsToFit(true);

      prevPlot = plot;

      return true;
      }
     
   private void positionPlotWindow() {
      IJ.wait(500);
      if (pwin==null || img==null) return;
      ImageWindow iwin = img.getWindow();
      if (iwin==null) return;
      Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
      Dimension plotSize = pwin.getSize();
      Dimension imageSize = iwin.getSize();
      if (plotSize.width==0 || imageSize.width==0) return;
      Point imageLoc = iwin.getLocation();
      int w = imageLoc.x+imageSize.width+10;
      if (w+plotSize.width>screen.width)
         w = screen.width-plotSize.width;
      pwin.setLocation(w, imageLoc.y);
      iwin.toFront();
      }

   private boolean updateProfile(int ix, int iy) {
      int nf = img.getNFrames();
      int cs = img.getZ();
      for(int f=0; f<nf; f++)
         y[f]  = img.getStack().getProcessor(img.getStackIndex(1,cs,f+1)).getPixelValue(ix, iy);

      Plot plot = new Plot("Frame Profile ("+img.getTitle()+")", xLabel, yLabel);
      plot.setFont(new Font("Comic Sans MS", Font.PLAIN, 20));
      plot.setXLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));
      plot.setYLabelFont(new Font("Comic Sans MS", Font.PLAIN, 24));

      if (prevPlot != null) plot.useTemplate(prevPlot, Plot.COPY_SIZE);

      plot.setColor(Color.blue);
      plot.setLineWidth(4);

      Calibration cal = img.getCalibration();
      plot.setJustification(Plot.RIGHT);
      plot.addLabel(0.99,0.99,String.format("%.2f(%d), %.2f(%d), (%d)",
                    cal.getX(ix),ix,cal.getY(iy),iy,img.getT()));

      plot.addPoints(x,y,Plot.X);
      if (pwin==null) {
         pwin = plot.show();
         pwin.addWindowListener(this);
         }
      else pwin.drawPlot(plot);

      prevPlot = plot;

      plot.setLimitsToFit(true);

      return true;
      }

   public void mousePressed(MouseEvent e)  { off(); if (updateProfile()) on(); }
   public void mouseDragged(MouseEvent e)  { off(); if (updateProfile()) on(); }
   public void keyReleased(KeyEvent e)     { off(); if (updateProfile()) on(); }

   public void keyPressed(KeyEvent e)      {}
   public void keyTyped(KeyEvent e)        {}
   public void mouseReleased(MouseEvent e) {}
   public void mouseExited(MouseEvent e)   {}
   public void mouseClicked(MouseEvent e)  {}   
   public void mouseEntered(MouseEvent e)  {}
   public void mouseMoved(MouseEvent e)    {}

   public void mouseWheelMoved(MouseWheelEvent e) { /* updateProfile(); */ }

   public void windowActivated(WindowEvent e)   {}
   public void windowClosed(WindowEvent e)      { disengage();}
   public void windowClosing(WindowEvent e)     { disengage();}
   public void windowDeactivated(WindowEvent e) {}
   public void windowDeiconified(WindowEvent e) {}
   public void windowIconified(WindowEvent e)   {}
   public void windowOpened(WindowEvent e)      {}
   }
